@@ -1,14 +1,19 @@
 import re
 import os
 import glob
-from typing import Literal, Tuple
+from typing import Literal, Tuple, Iterable
 import pyarrow.parquet as pq
 import xlrd
 import pandas as pd
 import numpy as np
 import pyarrow as pa
-from scipy.interpolate import NearestNDInterpolator
 from colorama import Fore, Back, Style
+from collections import Counter
+import string
+import copy
+from typing import Optional, List
+
+SILLY_DELIMITERS = frozenset(string.ascii_letters + string.digits + ".")
 
 
 def caster(df: pd.DataFrame):
@@ -54,13 +59,47 @@ def caster(df: pd.DataFrame):
 
 
 def load(
-    file: str,
+    file: str | Iterable[str],
     sheet: int = 1,
-    skiprows=None,
+    skiprows = None,
     on: str = "index",
-    classic_data=True,
-    recursive=True,
+    classic_data: bool = True,
+    recursive: bool = True,
+    has_title: bool = True,
+    has_index: bool = True
 ):
+    """Load files or folders for most used file types.
+
+    Allowed file extensions are :
+        - csv
+        - xlsx
+        - xls
+        - txt
+        - png
+        - jpg
+        - pkl
+        - bmp
+        - xlsm
+        - json
+        - parquet
+        - wav
+
+    Args:
+        file: The file or folder name.
+            If it is a folder, all supported file types will be
+            loaded in a list.
+        sheet:
+        skiprows:
+        on:
+        classic_data:
+        recursive:
+
+    """
+
+    if type(file) != str:
+        return [
+            load(elem, sheet, skiprows, on, classic_data, recursive) for elem in file
+        ]
     split_path = file.split("/")
     if len(split_path) == 1:
         directory = "./"
@@ -84,11 +123,12 @@ def load(
         )  # TODO check multi extension
         raise Exception
     elif len(files) == 0:
-        extract = pd.DataFrame()
         print("Warning : The file '" + file + "' wasn't found !")
-    else:
-        file = files[0].split(file)[0] + file
-        if file_format in ("xlsx", "xls", "xlsm"):
+        return pd.DataFrame()
+
+    file = files[0].split(file)[0] + file
+    match file_format:
+        case "xlsx" | "xls" | "xlsm":
             excel = pd.ExcelFile(file)
             sheets = excel.sheet_names
             try:
@@ -121,20 +161,20 @@ def load(
                 extract = caster(extract)
             extract.set_index(extract.columns[0])
             extract.drop(extract.columns[0], axis=1, inplace=True)
-        elif file_format == "pkl":
+        case "pkl":
             if not os.path.isfile(file):
                 print("The specified pickle file doesn't exist !")
             extract = pd.read_pickle(file)
-        elif file_format == "csv":
-            from detect_delimiter import detect
-
+        case "csv":
             with open(file) as myfile:
                 firstline = myfile.readline()
                 delimiter = detect(firstline, default=";")
                 myfile.close()
+            print(delimiter)
             extract = pd.read_csv(file, delimiter=delimiter)
-
-            extract.dropna(how="all", inplace=True)
+            extract.dropna(how="all", inplace=True) # Drop empty rows
+            if not has_title:
+                extract = extract.T.reset_index().T.reset_index(drop=True)
             if len(
                 [
                     True
@@ -148,11 +188,13 @@ def load(
                 )
             if classic_data:
                 if on != "index":
-                    extract.dropna(subset=[on], inplace=True)
+                    extract.dropna(subset=[on], inplace=True) # Drop rows without label
                 extract = caster(extract)
-        elif file_format == "parquet":
+            if has_index:
+                extract = extract.set_index(extract.columns[0]) # Set first column as index
+        case "parquet":
             extract = pq.read_pandas(file).to_pandas()
-        elif file_format in ["png", "jpg"]:
+        case "png" | "jpg":
             from cv2 import imread
 
             img = imread(file)
@@ -161,45 +203,51 @@ def load(
             ).all():
                 return img[:, :, 0]
             return img
-        elif file_format == "bmp":
+        case "bmp":
             import imageio
 
             return imageio.v3.imread(file)
-        elif file_format == "json":
+        case "json":
             import json
 
             with open(file) as user_file:
                 file_contents = user_file.read()
             return json.loads(file_contents)
-        elif file_format == "txt":
+        case "txt":
             with open(file) as f:
-                lines = f.readlines()
+                lines = f.read().splitlines()
             return lines
-        elif file_format == None:  # Directory
+        case "wav":
+            from librosa import load as librosa_load
+
+            return librosa_load(file, sr=None)
+        case None:  # Directory
             return [
                 load(elem, sheet, skiprows, on, classic_data, recursive)
                 for elem in glob.glob(file + "/*")
             ]
-        else:
-            print("I don't handle this file format yet, came back in a decade.")
+        case default:
+            print(
+                f"I don't handle this file format yet ({default}), came back in a decade."
+            )
             raise Exception
-        if classic_data:
-            if on == "index":
-                extract = extract[~extract.index.duplicated(keep="first")]
-            elif (not (on in extract)) and (on != None):
-                print(
-                    "There is no column name '"
-                    + on
-                    + "' in the sheet "
-                    + str(sheet)
-                    + " of the file '"
-                    + file
-                    + "' so I can't load this file? Try to change the 'on' parameter"
-                )
-                raise Exception
-            else:
-                extract.drop_duplicates(subset=on, keep="last", inplace=True)
-            extract.replace("Null", np.nan, inplace=True)
+    if classic_data:
+        if on == "index":
+            extract = extract[~extract.index.duplicated(keep="first")]
+        elif (not (on in extract)) and (on != None):
+            print(
+                "There is no column name '"
+                + on
+                + "' in the sheet "
+                + str(sheet)
+                + " of the file '"
+                + file
+                + "' so I can't load this file? Try to change the 'on' parameter"
+            )
+            raise Exception
+        else:
+            extract.drop_duplicates(subset=on, keep="last", inplace=True)
+        extract.replace("Null", np.nan, inplace=True)
     return extract
 
 
@@ -208,50 +256,74 @@ def save(
     file: str,
     sheet="Data",
     keep_index=True,
-):  # FIXME column width when save excel
-    file_format = file.split(".")[-1]
-    if file_format == "xlsx":
-        try:
-            writer = pd.ExcelWriter(file)
-        except PermissionError:
+):
+    """Save different object types to different file types.
+
+    Supported file types are:
+        - pkl
+        - parquet
+        - json
+        - xlsx
+        - png
+
+    Args:
+        file: The file name.
+        sheet: The sheet name the data is saved if saved in an excel.
+        keep_index: Keep the indexes in the saved file.
+
+    Raises:
+        Exception: If the file can't be accessed or the file type is
+        not supported.
+
+    """
+
+    match file.split(".")[-1]:
+        case "xlsx":
+            try:
+                writer = pd.ExcelWriter(file)
+            except PermissionError:
+                print(
+                    "I can't access the file '" + file + "', the "
+                    "solution may be to close it and retry."
+                )
+                raise Exception
+            try:
+                object_to_save.to_excel(writer, sheet_name=sheet, index=keep_index)
+            except IOError:
+                object_to_save.to_excel(writer, sheet_name=sheet, index=True)
+            # for column in object_to_save:
+            #    column_length = max(object_to_save[column].astype(str).map(len).max(), len(column))
+            #    col_idx = object_to_save.columns.get_loc(column)
+            #    writer.sheets[sheet].set_column(col_idx, col_idx, column_length)
+            writer._save()
+        case "csv":
+            object_to_save.to_csv("./" + file, index=keep_index)
+        case "pkl":
+            pd.to_pickle(object_to_save, "./" + file)
+        case "parquet":
+            caster(object_to_save)
+            table = pa.Table.from_pandas(object_to_save, preserve_index=False)
+            pq.write_table(table, file + ".parquet")
+        case "png":
+            from PIL import Image
+
+            im = Image.fromarray(object_to_save)
+            im.save(file)
+        case "json":
+            with open(file, "w") as outfile:
+                outfile.write(object_to_save)
+        case default:
             print(
-                "I can't access the file '" + file + "', the "
-                "solution may be to close it and retry."
+                f"I don't handle this file format yet ({default}), came back in a decade."
             )
             raise Exception
-        try:
-            object_to_save.to_excel(writer, sheet_name=sheet, index=keep_index)
-        except IOError:
-            object_to_save.to_excel(writer, sheet_name=sheet, index=True)
-        # for column in object_to_save:
-        #    column_length = max(object_to_save[column].astype(str).map(len).max(), len(column))
-        #    col_idx = object_to_save.columns.get_loc(column)
-        #    writer.sheets[sheet].set_column(col_idx, col_idx, column_length)
-        writer.save()
-    elif file_format == "pkl":
-        pd.to_pickle(object_to_save, "./" + file)
-    elif file_format == "parquet":
-        caster(object_to_save)
-        table = pa.Table.from_pandas(object_to_save, preserve_index=False)
-        pq.write_table(table, file + ".parquet")
-    elif file_format == "png":
-        from PIL import Image
-
-        im = Image.fromarray(object_to_save)
-        im.save(file)
-    elif file_format == "json":
-        with open(file, "w") as outfile:
-            outfile.write(object_to_save)
-    else:
-        print("I don't handle this file format yet, came back in a decade.")
-        raise Exception
 
 
 def merge(
     df: pd.DataFrame,
     file: str | pd.DataFrame,
     sheet=1,
-    on="index",
+    on: str = "index",
     skiprows=None,
     how: str = "outer",
     save_file: bool = False,
@@ -275,13 +347,15 @@ def merge(
                 columns.remove(on)
             except ValueError:
                 print(
-                    "You can't merge your data as there are not column '"
+                    "You can't merge your data as there are no column '"
                     + on
                     + "' in your already loaded DataFrame."
                 )
                 raise Exception
         if df.empty:
             merge = dataBase.copy()
+        elif on != "index":
+            merge = dataBase.merge(df, how=how, on=on)
         else:
             merge = dataBase.merge(df, how=how, left_index=True, right_index=True)
         merge = merge.loc[:, ~merge.columns.duplicated()]
@@ -307,98 +381,6 @@ def merge(
         return merge
 
 
-def interpolator(df: pd.DataFrame) -> pd.DataFrame:
-    array = np.array(df)
-    filled_array = array[~np.isnan(array).any(axis=1), :]
-    for i in range(array.shape[1]):
-        idxs = list(range(array.shape[1]))
-        idxs.pop(i)
-        my_interpolator = NearestNDInterpolator(
-            filled_array[:, idxs], filled_array[:, i], rescale=True
-        )
-        array[:, i] = np.apply_along_axis(
-            lambda row: my_interpolator(*row[idxs]) if np.isnan(row[i]) else row[i],
-            1,
-            array,
-        )
-    return caster(pd.DataFrame(array, columns=df.columns, index=df.index))
-
-
-def prepare_table(
-    df: pd.DataFrame,
-    y: str | list[str] = "",
-    correlation_threshold: float = 0.9,
-    missing_rows_threshold: float = 0.9,
-    missing_columns_threshold: float = 0.6,
-    categories_ratio_threshold: float = 0.1,
-    id_correlation_threshold: float = 0.04,
-    verbose: bool = False,
-) -> Tuple[pd.DataFrame, pd.DataFrame]:
-    """
-    Prepare the table to feed a ML model.
-    :param df: str, Optional
-    :param y:
-    :param correlation_threshold:
-    :param missing_rows_threshold:
-    :param missing_columns_threshold:
-    :param categories_ratio_threshold:
-    :param id_correlation_threshold:
-    :return:
-    """
-    y = [y] if not y is list else y
-    category_columns = df.select_dtypes(include=["category", object]).columns
-    # Drop rows with NaNs in tye y columns
-    if y != [""]:
-        for col in y:
-            df = df[df[col].notna()]
-    # Drop the categorical columns with too much categories
-    df = df.drop(
-        [
-            col
-            for col in category_columns
-            if df[col].nunique() / len(df) > categories_ratio_threshold
-        ],
-        axis=1,
-    )
-    # Convert categorical data to numerical ones
-    df = pd.get_dummies(df)
-    # Drop columns with not enough data
-    df = df.loc[
-        :,
-        df.apply(
-            lambda col: (1 - col.count() / len(df.index)) < missing_columns_threshold,
-            axis=0,
-        ),
-    ]
-    # Drop rows with not enough data
-    df = df[
-        df.apply(
-            lambda row: (1 - row.count() / len(df.columns)) < missing_rows_threshold,
-            axis=1,
-        )
-    ]
-    correlate_couples = []
-    corr = df.corr().abs()
-    for col in corr:
-        for index, val in corr[col].items():
-            if val > correlation_threshold and (index != col):
-                if (index, col) not in correlate_couples:
-                    correlate_couples.append((col, index))
-        if (df[col].nunique() == df.shape[0]) and (
-            (corr[col].sum() - 1) / corr.shape[0] < id_correlation_threshold
-        ):  # TODO améliorer
-            # Drop ids columns (unique values with low correlations with other columns)
-            if not col in y:
-                df = df.drop(col, axis=1)
-    for couple in correlate_couples:
-        # Drop a column if highly correlated with another one
-        if not any(elem in couple for elem in y):
-            df = df.drop(couple[0], axis=1)
-    if verbose:
-        print("")
-    return interpolator(df.drop(y, axis=1)), df[y]
-
-
 def color(
     text: str,
     fg: Literal[
@@ -418,7 +400,7 @@ def color(
         "LIGHTMAGENTA",
         "LIGHTCYAN",
         "LIGHTWHITE",
-    ],
+    ] = "RED",
     bg: Literal[
         "BLACK",
         "RED",
@@ -438,10 +420,128 @@ def color(
         "LIGHTWHITE",
     ] = None,
 ) -> str:
-    if "LIGHT" in bg:
+    """Format the text to print it in color.
+
+    Args:
+        text: The input text to format.
+        fg: The foreground color.
+        bg: The background color.
+
+    Examples:
+        >>> "This is a " + color("blue text", "BLUE") + " !"
+        'This is a \\x1b[34mblue text\\x1b[0m !'
+
+        >>> print("This is a " + color("blue text", "BLUE") + " !")
+        This is a \x1b[34mblue text\x1b[0m !
+
+        >>> "This is a " + color("strange color", "GREEN", "WHITE") + " !"
+        'This is a \\x1b[47m\\x1b[32mstrange color\\x1b[0m !'
+
+    """
+
+    if bg and ("LIGHT" in bg):
         bg += "_EX"
     if "LIGHT" in fg:
         fg += "_EX"
-    bg = Back.__getattribute__(bg)
     colored_text = f"{Fore.__getattribute__(fg)}{text}{Style.RESET_ALL}"
-    return colored_text if bg is None else bg + colored_text
+    return colored_text if bg is None else Back.__getattribute__(bg) + colored_text
+
+
+def _same(sequence):
+    elements = iter(sequence)
+    try:
+        first = next(elements)
+    except StopIteration:
+        return True
+    for el in elements:
+        if el != first:
+            return False
+    return True
+
+
+def detect(
+    text: str,
+    default: Optional[str] = None,
+    whitelist: Optional[List[str]] = None,
+    blacklist: Optional[List[str]] = SILLY_DELIMITERS,
+) -> Optional[str]:
+    r"""Detects the delimiter used in text formats such as CSV and its many counsins.
+
+    >>> detect(r"looks|like|the vertical bar\nis|the|delimiter\n")
+    '|'
+
+    `detect_delimiter.detect()` looks at the text provided to try to
+    find an uncommon delimiter, such as ` for whatever reason.
+
+    >>> detect('looks\x10like\x10something stupid\nis\x10the\x10delimiter')
+    '\x10'
+
+    When `detect()` doesn't know, it returns `None`:
+
+    >>> text = "not really any delimiters in here.\nthis is just text.\n"
+    >>> detect(text)
+
+    It's possible to provide a default, which will be used in that case:
+
+    >>> detect(text, default=',')
+    ','
+
+    By default, it will prevent avoid checking alpha-numeric characters
+    and the period/full stop character ("."). This can be adjusted via
+    the `blacklist` parameter.
+
+    If you believe that you know the delimiter, it's possible to provide
+    a list of possible delimiters to check for via the `whitelist` parameter.
+    If you don't provide a value, `[',', ';', ':', '|', '\t']` will be checked.
+    """
+
+    if whitelist:
+        candidates = whitelist
+    else:
+        candidates = list(",;:|\t")
+
+    sniffed_candidates = Counter()
+    likely_candidates = []
+
+    lines = []
+    # todo: support streaming
+    text_ = copy.copy(text)
+    while len(lines) < 5:
+        for line in text_.splitlines():
+            lines.append(line)
+
+    for c in candidates:
+        fields_for_candidate = []
+
+        for line in lines:
+            for char in line:
+                if char not in blacklist:
+                    sniffed_candidates[char] += 1
+            fields = line.split(c)
+            n_fields = len(fields)
+
+            # if the delimiter isn't present in the
+            # first line, it won't be present in the others
+            if n_fields == 1:
+                break
+            fields_for_candidate.append(n_fields)
+
+        if not fields_for_candidate:
+            continue
+
+        if _same(fields_for_candidate):
+            likely_candidates.append(c)
+
+    # no delimiter found
+    if not likely_candidates:
+        if whitelist is None and sniffed_candidates:
+            new_whitelist = [
+                char for (char, _count) in sniffed_candidates.most_common()
+            ]
+            return detect(text, whitelist=new_whitelist) or default
+        return default
+
+    if default in likely_candidates:
+        return default
+
+    return likely_candidates[0]
